/****************************************/
/* MES/Copyleft Yukio Mituiwa,2005	*/
/*					*/
/*  2006/6/16 first release		*/
/*  2006/6/30 fix size param		*/
/*					*/
/****************************************/
#include <module.h>
#include <string.h>

#ifdef _KERNEL_
  #include "mes.h"
  #include "conf.h"
  #define device_main(arg1,arg2) init_i2c(void)
  extern PIOConf i2c_config[];
#endif

typedef struct {
	IOREG		*regptr;
	int		port, size, offset, blksiz;
	unsigned char	bits[2];
} I2CConf;
#define I2CNUM	2
static I2CConf i2c[I2CNUM];

#define SCL	(i2c[minor].bits[0])
#define SDA	(i2c[minor].bits[1])
#define DR	*(i2c[minor].regptr)
#define A1WRITE	0xa2
#define A1READ	0xa3
#define WRITE	0xa0
#define READ	0xa1
#define NAK	SDA

static void i2c_start(int minor) {
	ddr_set(i2c[minor].port, SDA);
	DR &= ~SCL;
	DR |= SDA;
	DR |= SCL;
	DR &= ~SDA;
	DR &= ~SCL;
	ddr_reset(i2c[minor].port, SDA);
}

static void i2c_stop(int minor) {
	ddr_set(i2c[minor].port, SDA);
	DR &= ~SCL;
	DR &= ~SDA;
	DR |= SCL;
	DR |= SDA;
	DR &= ~SCL;
	ddr_reset(i2c[minor].port, SDA);
}

static int i2c_write(int minor, int data) {
	int	i, mask;

	ddr_set(i2c[minor].port, SDA);
	mask = 0x80;
	for(i = 0;i < 8;i++) {
		if(data & mask) DR |= SDA;
		else DR &= ~SDA;
		DR |= SCL;
		DR &= ~SCL;
		mask >>= 1;
	}
	ddr_reset(i2c[minor].port, SDA);
	DR |= SCL;
	i = DR & SDA;
	DR &= ~SCL;
	return i;
}


static int i2c_read(int minor) {
	int	i, data;

	data = 0;
	for(i = 0;i < 8;i++) {
		data <<= 1;
		DR |= SCL;
		if(DR & SDA) data++;
		DR &= ~SCL;
	}
	return data;
}

static void i2c_ack(int minor) {
	ddr_set(i2c[minor].port, SDA);
	DR &= ~SDA;
	DR |= SCL;
	DR &= ~SCL;
	ddr_reset(i2c[minor].port, SDA);
}

static void i2c_nak(int minor) {
	ddr_set(i2c[minor].port, SDA);
	DR |= SDA;
	DR |= SCL;
	DR &= ~SCL;
	ddr_reset(i2c[minor].port, SDA);
}

static void send_ctrl(int minor, int ctrl) {
	int	ret;

	do {
		i2c_start(minor);
		ret = i2c_write(minor, ctrl);
	} while(ret == NAK);
}

static void write_prepare(int minor, int addr) {
	if(addr & 0x10000) send_ctrl(minor, A1WRITE);
	else send_ctrl(minor, WRITE);
	i2c_write(minor, (addr >> 8) & 0xff);
	i2c_write(minor, addr & 0xff);
}

static void read_prepare(int minor, int addr) {
	write_prepare(minor, addr);
	if(addr & 0x10000) send_ctrl(minor, A1READ);
	else send_ctrl(minor, READ);
}

static int write_i2c(int minor, char *src, int size) {
	int	i;

	if(minor >= I2CNUM) return -1;
	if(i2c[minor].blksiz == 0) return -1;
	if(i2c[minor].offset >= i2c[minor].size && i2c[minor].blksiz > 1) return -1;
	if(size == 0) return 0;
	write_prepare(minor, i2c[minor].offset);
	for(i = 0;i < size - 1;i++) {
		i2c_write(minor, src[i]);
		if((i % i2c[minor].blksiz) == (i2c[minor].blksiz - 1)) {
			i2c_stop(minor);
			write_prepare(minor, i2c[minor].offset + i + 1);
		}
	}
	i2c_write(minor, src[i]);
	i2c_stop(minor);
	return size;
}

static int read_i2c(int minor, char *dest, int size) {
	int	i;

	if(minor >= I2CNUM) return -1;
	if(i2c[minor].blksiz == 0) return -1;
	if(i2c[minor].offset >= i2c[minor].size && i2c[minor].blksiz > 1) return -1;
	if(size == 0) return 0;
	read_prepare(minor, i2c[minor].offset);
	for(i = 0;i < size - 1;i++) {
		dest[i] = i2c_read(minor);
		if((i % i2c[minor].blksiz) == (i2c[minor].blksiz - 1)) {
			i2c_nak(minor);
			i2c_stop(minor);
			read_prepare(minor, i2c[minor].offset + i + 1);
		} else {
			i2c_ack(minor);
		}
	}
	dest[i] = i2c_read(minor);
	i2c_nak(minor);
	i2c_stop(minor);
	return size;
}

static int open_i2c(int minor, int option) {
	if(minor >= I2CNUM) return -1;
	ddr_reset(i2c[minor].port, SDA);
	ddr_set(i2c[minor].port, SCL);
	DR &= ~SCL;
	return 0;
}

static int close_i2c(int minor) {
	if(minor >= I2CNUM) return -1;
	if(i2c[minor].blksiz == 0) return -1;
	return 0;
}

static int seek_i2c(int minor, int pos) {
	if(minor >= I2CNUM) return -1;
	if(i2c[minor].blksiz == 0) return -1;
	i2c[minor].offset = pos;
	return 0;
}

static int ioctl_i2c(int minor, int data, int op) {
	int	ret;

	if(minor >= I2CNUM) return -1;
	if(i2c[minor].blksiz == 0) return -1;
	ret = 0;
	switch(op) {
	case DEV_INFO:
		break;
	case MEDIA_SIZE:
		ret = i2c[minor].size;
		break;
	default:
		ret = -1;
	}
	return ret;
}

static Functions func;

int device_main(int argc, char **argv) {
	unsigned int	minor;

#ifndef _KERNEL_
	minor = 0;
	for(n = 1;n < argc;n++) {
		if(memcmp(argv[n], "i2c", 3) == 0) {
			minor = stoi(&argv[n][3]);
			if(minor >= I2CNUM) minor = 0;
		} else if(memcmp(argv[n], "p=", 2) == 0) {
			i2c[minor].port = stoi(&argv[n][2]);
		} else if(memcmp(argv[n], "clk=", 4) == 0) {
			i2c[minor].bits[0] = 1 << stoi(&argv[n][4]);
		} else if(memcmp(argv[n], "dat=", 4) == 0) {
			i2c[minor].bits[1] = 1 << stoi(&argv[n][4]);
		} else if(memcmp(argv[n], "size=", 5) == 0) {
			i2c[minor].size = stoh(&argv[n][5]);
		}
	}
#endif

	for(minor = 0;minor < I2CNUM;minor++) {
#ifdef _KERNEL_
		i2c[minor].port = i2c_config[minor].port;
		i2c[minor].bits[0] = 1 << i2c_config[minor].bits[1];
		i2c[minor].bits[1] = 1 << i2c_config[minor].bits[3];
		i2c[minor].size = i2c_config[minor].size;
#endif
		i2c[minor].regptr = get_dr(i2c[minor].port);
		if(i2c[minor].size > 0x10000) {
			i2c[minor].blksiz = 256;
		} else if(i2c[minor].size > 0x8000) {
			i2c[minor].blksiz = 128;
		} else if(i2c[minor].size > 0x10) {
			i2c[minor].blksiz = 64;
		} else {
			i2c[minor].blksiz = 1;
		}
		i2c[minor].offset = 0;
	}
	func.open_dev = open_i2c;
	func.close_dev = close_i2c;
	func.write_dev = write_i2c;
	func.read_dev = read_i2c;
	func.seek_dev = seek_i2c;
	func.ioctl_dev = ioctl_i2c;
	func.poll_dev = 0;
	strcpy(&(func.name[0]), "i2c");
	device_return(&func);
}
